# Generated by Django 2.0.13 on 2019-12-29 00:20
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("webauthin", "0003_auto_20191228_1750")]

    operations = [
        migrations.AlterModelOptions(
            name="authdata", options={"verbose_name_plural": "auth data"}
        )
    ]



import base64

import webauthn
from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site

from .models import AuthData

ICON = "https://example.com/"


class WebAuthinBackend:
    def get_user(self, user_id):
        """Get a user by their primary key."""
        User = get_user_model()
        try:
            return User.objects.get(pk=user_id)
        except User.DoesNotExist:
            return None

    def authenticate(self, request, credential_id, data=None):
        """Authenticate a user given a signed token."""
        challenge = request.session.get("challenge")
        if not challenge:
            return None

        auth_data = AuthData.objects.filter(credential_id=data["id"]).first()
        if not auth_data:
            return None

        site = get_current_site(request)

        webauthn_user = webauthn.WebAuthnUser(
            user_id=base64.b64encode(str(auth_data.user.id).encode()).decode(),
            username=auth_data.user.get_username(),
            display_name="User",
            icon_url=ICON,
            credential_id=auth_data.credential_id,
            public_key=auth_data.public_key,
            sign_count=auth_data.sign_count,
            rp_id=site.domain,
        )

        webauthn_assertion_response = webauthn.WebAuthnAssertionResponse(
            webauthn_user, data, challenge, "https://%s" % site.domain, uv_required=True
        )

        try:
            sign_count = webauthn_assertion_response.verify()
        except Exception:
            return None

        # TODO: Validate the sign count here (https://w3c.github.io/webauthn/#signature-counter)
        auth_data.set_sign_count(sign_count)
        return auth_data.user



from django.contrib import admin

from .models import AuthData


@admin.register(AuthData)
class AuthDataAdmin(admin.ModelAdmin):
    list_display = ["user", "credential_id", "created_on", "last_used_on"]
    search_fields = ["credential_id"]



# Generated by Django 2.0.13 on 2019-12-28 17:50
import datetime

import django.utils.timezone
from django.db import migrations
from django.db import models
from django.utils.timezone import utc


class Migration(migrations.Migration):

    dependencies = [("webauthin", "0002_authdata_sign_count")]

    operations = [
        migrations.AddField(
            model_name="authdata",
            name="created_on",
            field=models.DateTimeField(
                auto_now_add=True,
                default=datetime.datetime(2019, 12, 28, 17, 50, 48, 20406, tzinfo=utc),
            ),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name="authdata",
            name="last_used_on",
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]



# Generated by Django 2.0.13 on 2019-12-28 15:44
import django.db.models.deletion
from django.conf import settings
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="AuthData",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("credential_id", models.CharField(max_length=300, unique=True)),
                ("public_key", models.CharField(max_length=500)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        )
    ]



import base64
import random
import string

import webauthn
from django.contrib import messages
from django.contrib.auth import authenticate
from django.contrib.auth import get_user_model
from django.contrib.auth import login
from django.contrib.auth.decorators import login_required
from django.contrib.sites.shortcuts import get_current_site
from django.http import JsonResponse
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt

from .models import AuthData

ICON = "https://example.com/"

User = get_user_model()


def get_challenge() -> str:
    return "".join(
        [
            random.SystemRandom().choice(string.ascii_letters + string.digits)
            for i in range(32)
        ]
    )


def login_view(request):
    return render(request, "webauthin_login.html")


@csrf_exempt
@login_required
def register_begin(request):
    site = get_current_site(request)
    challenge = get_challenge()
    request.session["challenge"] = challenge
    reg_data = webauthn.WebAuthnMakeCredentialOptions(
        challenge=challenge,
        rp_name=site.name,
        rp_id=site.domain,
        user_id=base64.b64encode(str(request.user.id).encode()).decode(),
        username=request.user.get_username(),
        display_name="%s user: %s" % (site.name, request.user.get_username()),
        icon_url=ICON,
        attestation="none",
        user_verification="required",
    ).registration_dict
    # This is needed so the key gets stored on the authenticator, I think.
    reg_data["authenticatorSelection"]["requireResidentKey"] = True
    return JsonResponse(reg_data)


@csrf_exempt
@login_required
def register_verify(request):
    site = get_current_site(request)
    challenge = request.session.get("challenge")
    if not challenge:
        return JsonResponse("No challenge exists in your session.", status=422)

    registration_response = request.POST
    webauthn_registration_response = webauthn.WebAuthnRegistrationResponse(
        rp_id=site.domain,
        origin="https://%s" % site.domain,
        registration_response=registration_response,
        challenge=challenge,
        self_attestation_permitted=True,
        none_attestation_permitted=True,
        uv_required=True,
    )

    try:
        webauthn_credential = webauthn_registration_response.verify()
    except Exception as e:
        messages.error(request, "Registration failed. Error: {}".format(e))
        return JsonResponse({"fail": "Registration failed. Error: {}".format(e)})

    auth_data = AuthData.objects.filter(credential_id=webauthn_credential.credential_id)
    if auth_data.exists():
        messages.error(
            request,
            "This key is already registered to an account. Try logging in with it.",
        )
        return JsonResponse({"fail": "Credential ID already exists."}, status=422)

    webauthn_credential.credential_id = str(webauthn_credential.credential_id, "utf-8")
    webauthn_credential.public_key = str(webauthn_credential.public_key, "utf-8")
    AuthData.objects.create(
        user=request.user,
        credential_id=webauthn_credential.credential_id,
        public_key=webauthn_credential.public_key,
    )
    messages.success(request, "Your key has been successfully registered.")
    return JsonResponse({"success": "User successfully registered."})


@csrf_exempt
def login_begin(request):
    site = get_current_site(request)
    challenge = get_challenge()
    request.session["challenge"] = challenge
    login_data = {
        "challenge": request.session["challenge"],
        "timeout": 60000,
        "rpId": site.domain,
        "allowCredentials": [],
        "userVerification": "required",
        "extensions": {
            "txAuthSimple": "FIDO",
            "txAuthGenericArg": {"contentType": "text/plain", "content": "RklETw=="},
            "uvi": True,
        },
        "status": "ok",
        "errorMessage": "Error while logging in.",
    }

    return JsonResponse(login_data)


@csrf_exempt
def login_verify(request):
    challenge = request.session.get("challenge")
    if not challenge:
        return JsonResponse("No challenge exists in your session.", status=422)

    print(request.POST)

    user = authenticate(request, credential_id=request.POST["id"], data=request.POST)
    if user is None:
        return JsonResponse({"fail": "Assertion failed."}, status=422)

    login(request, user)

    return JsonResponse({"success": True})



from django.contrib.auth import get_user_model
from django.db import models
from django.utils.timezone import now


class AuthData(models.Model):
    user = models.ForeignKey(get_user_model(), on_delete=models.CASCADE)
    credential_id = models.CharField(max_length=300, unique=True)
    public_key = models.CharField(max_length=500)
    sign_count = models.IntegerField(default=0)
    created_on = models.DateTimeField(auto_now_add=True)
    last_used_on = models.DateTimeField(default=now)

    class Meta:
        verbose_name_plural = "auth data"

    def set_sign_count(self, sign_count: int) -> None:
        """Set the sign_count and update the last_used_on datetime."""
        self.sign_count = sign_count
        self.last_used_on = now()
        self.save()

    def __str__(self):
        return str(self.user)



# Generated by Django 2.0.13 on 2019-12-28 16:05
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [("webauthin", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="authdata",
            name="sign_count",
            field=models.IntegerField(default=0),
        )
    ]



from django.conf import settings

# The rate at which people will be able to request logins.
RATELIMIT_RATE = getattr(settings, "TOKENAUTH_RATELIMIT_RATE", "3/h")

NORMALIZE_EMAIL = getattr(settings, "TOKENAUTH_NORMALIZE_EMAIL", lambda e: e)

# How long a token should be valid for, in seconds.
TOKEN_DURATION = getattr(settings, "TOKENAUTH_TOKEN_DURATION", 30 * 60)

# How long the token should be, in characters.
TOKEN_LENGTH = getattr(settings, "TOKENAUTH_TOKEN_LENGTH", 8)

# Where to redirect after link operations.
LOGIN_URL = getattr(settings, "TOKENAUTH_LOGIN_URL", settings.LOGIN_URL)

# Where to redirect after login or logout.
LOGIN_REDIRECT = getattr(
    settings, "TOKENAUTH_LOGIN_REDIRECT", settings.LOGIN_REDIRECT_URL
)
LOGOUT_REDIRECT = getattr(
    settings,
    "TOKENAUTH_LOGOUT_REDIRECT",
    settings.LOGOUT_REDIRECT_URL or settings.LOGIN_REDIRECT_URL,
)

DEFAULT_FROM_EMAIL = getattr(
    settings, "TOKENAUTH_DEFAULT_FROM_EMAIL", settings.DEFAULT_FROM_EMAIL
)

SINGLE_USE_LINK = getattr(settings, "TOKENAUTH_SINGLE_USE_LINK", False)



from django.urls import path

from . import views


app_name = "webauthin"
urlpatterns = [
    path(r"login/begin/", views.login_begin, name="login-begin"),
    path(r"login/verify/", views.login_verify, name="login-verify"),
    path(r"register/begin/", views.register_begin, name="register-begin"),
    path(r"register/verify/", views.register_verify, name="register-verify"),
    path(r"login/", views.login_view, name="login"),
]



#!/usr/bin/env python
import sys

from webauthin import __version__

assert sys.version >= "3.6", "Requires Python v2.7 or above."
from distutils.core import setup  # noqa
from setuptools import find_packages  # noqa

setup(
    name="django-webauthin",
    version=__version__,
    author="Stavros Korokithakis",
    author_email="hi@stavros.io",
    url="https://gitlab.com/stavros/django-webauthin",
    description="""Passwordless authentication using WebAuthn.""",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    license="BSD",
    keywords="django",
    zip_safe=False,
    include_package_data=True,
    install_requires=["webauthn"],
    packages=find_packages(),
    python_requires=">=3.6",
)
